@@ -9,6 +9,7 @@ from .plots import *
 from . import utils, __version__, cpus_available
 
 import os
+import json
 import logging
 import subprocess
 import collections
@@ -19,8 +20,12 @@ from typing import Optional, Mapping
 from tqdm import tqdm
 import click
 import pandas as pd
+import matplotlib
 
+# For CLI, use the Agg background which doesn't support plt.show() and is faster
+matplotlib.use('Agg')
 
+# adds both -h and --help as options
 CONTEXT_SETTINGS = {'help_option_names': ['-h', '--help']}
 
 class _OrderedGroup(click.Group):
@@ -57,7 +62,8 @@ def cli():
     help="Optional delimited text file with metadata for features with one row each. Columns are annotation layers.")
 @click.option(
     "--sparsify", is_flag=True,
-    help="Save resulting data in sparse format. Recommended to increase performance for sparse datasets such as scRNA-Seq, scATAC-Seq, and 10X Visium, but not for bulk expression data.")
+    help="Save resulting data in sparse format. Recommended to increase performance for sparse datasets such as scRNA-Seq, scATAC-Seq, and 10X Visium"
+         ", but not for bulk expression data. [Note that this feature is experimental]")
 @click.option(
     "--transpose", is_flag=True,
     help="Transpose an input data matrix where rows are genes/features and columns are samples/cells into the correct orientation.")
@@ -190,6 +196,7 @@ def cmd_check_h5ad(input, output):
     """
     utils.start_logging()
     dataset = Dataset.from_h5ad(input)
+    dataset.remove_unfactorizable_observations()
     dataset.remove_unfactorizable_features()
     
     # Save output to new h5ad file
@@ -238,7 +245,8 @@ def cmd_model_odg(name, output_dir, input, default_spline_degree, default_dof, m
         mosaicmpi model-odg -n test -i test.h5ad --default_spline_degree 0 --default_dof 1
     """
     cNMF(output_dir=output_dir, name=name)  # creates directories for cNMF
-    utils.start_logging(os.path.join(output_dir, name, "logfile.txt"))
+    os.makedirs(os.path.join(output_dir, name, "logs"), exist_ok=True)
+    utils.start_logging(os.path.join(output_dir, name, "logs", "logfile.txt"))
     dataset = Dataset.from_h5ad(input)
     
     # Create gene stats table and save h5ad file
@@ -254,18 +262,13 @@ def cmd_model_odg(name, output_dir, input, default_spline_degree, default_dof, m
 
     # create mean vs variance plots
     fig = plot_feature_dispersion(dataset, show_selected=False)
-    fig.savefig(os.path.join(output_dir, name, "feature_meanvar.pdf"), facecolor='white')
-    fig.savefig(os.path.join(output_dir, name, "feature_meanvar.png"), dpi=400, facecolor='white')
+    utils.save_fig(fig, os.path.join(output_dir, name, "feature_meanvar"), formats=("pdf", "png"), target_dpi=400, facecolor='white')
     
-
     if dataset.is_imputed:
         logging.info("Creating plots for imputed data")
 
         fig = plot_feature_missingness(dataset, proportion=True)
-        fig.savefig(os.path.join(output_dir, name, "missingness_histogram.pdf"), facecolor='white')
-        fig.savefig(os.path.join(output_dir, name, "missingness_histogram.png"), dpi=400, facecolor='white')
-
-        fig = plot_feature_dispersion(dataset)
+        utils.save_fig(fig, os.path.join(output_dir, name, "missingness_histogram"), formats=("pdf", "png"), target_dpi=400, facecolor='white')
 
     logging.info("All tasks completed successfully.")
 
@@ -335,8 +338,8 @@ def cmd_set_parameters(name, output_dir, odg_method, odg_param, min_mean, k_rang
         # input a gene list from text file
         mosaicmpi set_parameters -n test -m genes_file -p path/to/genesfile.txt
     """
-    os.makedirs(os.path.join(output_dir, name), exist_ok=True)
-    utils.start_logging(os.path.join(output_dir, name, "logfile.txt"))
+    os.makedirs(os.path.join(output_dir, name, "logs"), exist_ok=True)
+    utils.start_logging(os.path.join(output_dir, name, "logs", "logfile.txt"))
     dataset = Dataset.from_h5ad(os.path.join(output_dir, name, name + ".h5ad"))
 
     if odg_method == "genes_file":
@@ -368,8 +371,7 @@ def cmd_set_parameters(name, output_dir, odg_method, odg_param, min_mean, k_rang
 
     # create mean vs variance plot, updated with selected genes
     fig = plot_feature_dispersion(dataset, show_selected=True)
-    fig.savefig(os.path.join(output_dir, name, "odgenes.pdf"), facecolor='white')
-    fig.savefig(os.path.join(output_dir, name, "odgenes.png"), dpi=400, facecolor='white')
+    utils.save_fig(fig, os.path.join(output_dir, name, "odgenes"), formats=("pdf", "png"), target_dpi=400, facecolor='white')
 
     # output table with gene overdispersion measures
     dataset.adata.var.to_csv(os.path.join(output_dir, name, "odgenes", "genestats.tsv"), sep="\t")
@@ -409,7 +411,8 @@ def cmd_factorize(name, output_dir, worker_index, total_workers, slurm_script):
     Performs factorization according to parameters specified using `mosaicmpi set-parameters`.
     """
     cnmf_obj = cNMF(output_dir=output_dir, name=name)
-    utils.start_logging(os.path.join(output_dir, name, "logfile.txt"))
+    os.makedirs(os.path.join(output_dir, name, "logs", "factorize"), exist_ok=True)
+    utils.start_logging(os.path.join(output_dir, name, "logs", "factorize", f"logfile_{worker_index}.txt"))
     
     run_params = utils.load_df_from_npz(cnmf_obj.paths['nmf_replicate_parameters'])
     if run_params.shape[0] == 0:
@@ -460,14 +463,14 @@ def cmd_postprocess(name, output_dir, cpus, local_density_threshold, local_neigh
                           os.getcwd(),
                           output_dir,
                           name,
-                          cpus,
-                          local_density_threshold,
-                          local_neighborhood_size,
-                          skip_missing_iterations,
-                          force_h5ad_update])
+                          str(cpus),
+                          str(local_density_threshold),
+                          str(local_neighborhood_size),
+                          flags])
     else:
         cnmf_obj = cNMF(output_dir=output_dir, name=name)
-        utils.start_logging(os.path.join(output_dir, name, "logfile.txt"))
+        os.makedirs(os.path.join(output_dir, name, "logs"), exist_ok=True)
+        utils.start_logging(os.path.join(output_dir, name, "logs", "logfile.txt"))
         cnmf_obj.postprocess(cpus=cpus,
                             local_density_threshold=local_density_threshold,
                             local_neighborhood_size=local_neighborhood_size,
@@ -527,8 +530,8 @@ def cmd_annotated_heatmap(input_h5ad, output_dir, metadata_colors_toml, max_cate
     
     # plot legend
     fig = colors.plot_metadata_colors_legend()
-    fig.savefig(os.path.join(output_dir, f"metadata_legend.pdf"))
-    
+    utils.save_fig(fig, os.path.join(output_dir, "metadata_legend"),)
+
     # filter metadata layers with too many categories
     if max_categories_per_layer is not None:
         subset_columns = dataset.get_metadata_df(include_numerical=False).apply(lambda x: len(x.cat.categories)) <= max_categories_per_layer
@@ -550,10 +553,44 @@ def cmd_annotated_heatmap(input_h5ad, output_dir, metadata_colors_toml, max_cate
             dataset=dataset, k=k, subset_metadata=subset_columns, colors=colors, title=title,
             cluster_samples=True, cluster_programs=False, show_sample_labels=(not hide_sample_labels))
         fig.savefig(filename, transparent=False, bbox_inches = "tight")
+        plt.close(fig)
     
     logging.info("All tasks completed successfully.")
 
 
+@click.command(name="map-gene-ids")
+@click.option('-i', '--input_h5ad', type=click.Path(exists=True, dir_okay=False), required=True, help="Input .h5ad file")
+@click.option('-o', '--output_h5ad', type=click.Path(exists=False, dir_okay=False), required=True, help="Output .h5ad file with mapped gene identifiers.")
+@click.option('--source_ids', type=click.Choice(["ensembl_gene", "gene_name"]), default="gene_name",
+              help="Whether the source feature IDs are gene names (eg., EGFR), or Ensembl genes (eg., ENSG00000146648)")
+@click.option('--dest_ids', type=click.Choice(["ensembl_gene", "gene_name"]), default="gene_name",
+              help="Whether the dest feature IDs are gene names (eg., EGFR), or Ensembl genes (eg., ENSG00000146648)")
+@click.option('--source_species', type=click.Choice(["hsapiens", "mmusculus", "rnorvegicus", "sscrofa", "dmelanogaster", "drerio", "celegans"]), required=True,
+              help="Species of source feature IDs")
+@click.option('--dest_species', type=click.Choice(["hsapiens", "mmusculus", "rnorvegicus", "sscrofa", "dmelanogaster", "drerio", "celegans"]), required=True,
+              help="Species of dest feature IDs")
+@click.option('--unmapped_prefix', type=str, default="unmapped_", help="String to prefix unmapped feature/gene IDs")
+@click.option('--one_to_many', is_flag=True, help="Map one-to-many relationships, duplicating features to accommodate.")
+def cmd_map_gene_ids(input_h5ad, output_h5ad, source_ids, dest_ids, source_species, dest_species, unmapped_prefix, one_to_many):
+    """
+    Map gene IDs for a dataset, keeping both mapped and unmapped IDs. By default, only one-to-one relationships are mapped.
+    """
+    utils.start_logging()
+    dataset = Dataset.from_h5ad(input_h5ad)
+    dataset.map_gene_ids(source_species=source_species,
+                         dest_species=dest_species,
+                         source_ids=source_ids,
+                         dest_ids=dest_ids,
+                         unmapped_prefix=unmapped_prefix,
+                         one_to_many=("duplicate" if one_to_many else False)
+                         )
+    message = "Feature counts by mapping relationship:"
+    for mapping_type, count in dataset.adata.var["mapping_relationship"].value_counts().items():
+        message += f"\n\t{mapping_type}: {count}"
+    logging.info(message)
+    dataset.write_h5ad(output_h5ad)
+    logging.info("All tasks completed successfully.")
+
 @click.command(name="create-config")
 @click.option('-i', '--input_h5ad', type=click.Path(exists=True, dir_okay=False), multiple=True, help=".h5ad file with cNMF results. Can be used multiple times to specify one or more datasets from which to create a config.toml file.")
 @click.option('-o', '--output_toml', type=click.Path(exists=False, dir_okay=False), required=False, help="Output .toml file for configuring integration.")
@@ -614,8 +651,19 @@ def cmd_integrate(output_dir, config_toml, communities_toml, colors_toml, cpus):
             
         if "patient_id_col" in dsparams:
             dataset.patient_id_col = dsparams["patient_id_col"]
+
         datasets[dsname] = dataset
 
+    # check datasets for duplicated features
+    datasets_dupfeatures = []
+    for dsname, dataset in datasets.items():
+        if not dataset.adata.var_names.unique:
+            datasets_dupfeatures.append(dsname)
+    if datasets_dupfeatures:
+        logging.error("Datasets with duplicated features cannot be used for integration with other datasets. "
+                      f"These datasets have duplicated features: {(', '.join(datasets_dupfeatures))}")
+        sys.exit(1)
+
     # gets k-values from config
     has_k_subset = {dsname: ("k_subset" in dsparams) for dsname, dsparams in config.datasets.items()}
     if all(list(has_k_subset.values())):
@@ -636,6 +684,14 @@ def cmd_integrate(output_dir, config_toml, communities_toml, colors_toml, cpus):
         k_subset = k_subset
         )
 
+    logging.info("Analyzing feature overlaps")
+    # Features and overdispersed features tables and UpSet plots
+    df = integration.get_overdispersed_features_overlap_table()
+    df.to_csv(os.path.join(output_dir, "overdispersed_features.txt"), sep="\t")
+
+    df = integration.get_features_overlap_table()
+    df.to_csv(os.path.join(output_dir, "features.txt"), sep="\t")
+    
     # save correlation matrix
     corr_path = os.path.join(output_dir, config.corr_method + ".df.npz")
     utils.save_df_to_npz(integration.corr_matrix, corr_path)
@@ -657,14 +713,6 @@ def cmd_integrate(output_dir, config_toml, communities_toml, colors_toml, cpus):
     fig = plot_pairwise_corr_overlaid(integration)
     utils.save_fig(fig, os.path.join(output_dir, f"pairwise_corr_overlaid"), target_dpi=300, formats=config.plot_formats)
 
-    logging.info("Analyzing feature overlaps")
-    # Features and overdispersed features tables and UpSet plots
-    df = integration.get_overdispersed_features_overlap_table()
-    df.to_csv(os.path.join(output_dir, "overdispersed_features.txt"), sep="\t")
-
-    df = integration.get_features_overlap_table()
-    df.to_csv(os.path.join(output_dir, "features.txt"), sep="\t")
-    
     if integration.n_datasets > 1:
         fig = plot_overdispersed_features_upset(integration)
         utils.save_fig(fig, os.path.join(output_dir, "overdispersed_features_upsetplot"), target_dpi=200, formats=config.plot_formats)
@@ -897,7 +945,7 @@ def cmd_integrate(output_dir, config_toml, communities_toml, colors_toml, cpus):
     logging.info("All tasks completed successfully.")
     
 @click.command(name="ssgsea")
-@click.option('-o', '--output_dir', type=click.Path(file_okay=False), required=True,
+@click.option('-o', '--output_dir', type=click.Path(file_okay=False), required=True, show_default=True,
     help="Output directory for ssgsea results")
 @click.option('-n', '--pkl_file', type=click.Path(exists=True, dir_okay=False),
     help="Path to network_integration.pkl.gz file from `mosaicmpi integrate` step")
@@ -905,9 +953,9 @@ def cmd_integrate(output_dir, config_toml, communities_toml, colors_toml, cpus):
     help="Path to .h5ad file from `mosaicmpi postprocess`")
 @click.option('-g', '--gene_sets', type=str, required=True, 
     help="Path to GMT file with gene sets or Enrichr Library name.")
-@click.option("--min_intersection", type=int, default=5, 
+@click.option("--min_intersection", type=int, default=5, show_default=True, 
     help="Minimum intersection size for gene sets")
-@click.option("--max_intersection", type=int, default=500, 
+@click.option("--max_intersection", type=int, default=500, show_default=True, 
     help="Minimum intersection size for gene sets")
 @click.option('--cpus', type=int, default=cpus_available, show_default=True,
     help="Number of CPUs for MP-enabled tasks")
@@ -941,7 +989,7 @@ def cmd_ssgsea(output_dir, pkl_file, h5ad_file, gene_sets, min_intersection, max
 
     if h5ad_file and not pkl_file:
         # run ssGSEA on all programs from a .h5ad file
-        dataset = Dataset.from_h5ad("rna.h5ad")
+        dataset = Dataset.from_h5ad(h5ad_file)
         programs = dataset.get_programs()
         result = gseapy.ssgsea(data=programs, gene_sets=gene_sets, min_size=min_intersection, max_size=max_intersection, threads=cpus)
         prog_nes = result.res2d.pivot(index="Term", columns="Name", values="NES")
@@ -969,6 +1017,162 @@ def cmd_ssgsea(output_dir, pkl_file, h5ad_file, gene_sets, min_intersection, max
         sys.exit(1)
     logging.info("All tasks completed successfully.")
     
+    
+@click.command(name="gprofiler")
+@click.option('-o', '--output_dir', type=click.Path(file_okay=False), required=True, show_default=True,
+    help="Output directory for gprofiler results")
+@click.option('-n', '--pkl_file', type=click.Path(exists=True, dir_okay=False),
+    help="Path to network_integration.pkl.gz file from `mosaicmpi integrate` step")
+@click.option('-i', '--h5ad_file', type=click.Path(exists=True, dir_okay=False),
+    help="Path to .h5ad file from `mosaicmpi postprocess`")
+@click.option('-g', '--gene_sets',
+              type = click.Choice(["GO:MF", "GO:CC", "GO:BP", "KEGG", "REAC", "WP", "TF", "MIRNA", "HPA", "CORUM", "HP"
+                                   ]),
+              multiple=True, default=[],
+    help="Source for gene sets from g:Profiler. Defaults to all sources. "
+    "Use multiple times to specify multiple sources, e.g.: -g GO:CC -g GO:BP")
+@click.option('-s', '--species', type=click.Choice(["hsapiens", "mmusculus"]), required=True,
+    help="Species for gene name IDs")
+@click.option("--min_intersection", type=int, default=5, show_default=True, 
+    help="Minimum intersection size for gene sets")
+@click.option("--max_intersection", type=int, default=500, show_default=True, 
+    help="Minimum intersection size for gene sets")
+@click.option("--n_hsg", type=int, default=1000, show_default=True)
+@click.option("--cmap", type=str, default="Blues",
+              help="matplotlib colormap name for heatmap plots.")
+@click.option("--vmin", type=float, default=0,
+              help="minimum -log10(pval) for heatmap plots")
+@click.option("--vmax", type=float, default=10,
+              help="maximum -log10(pval) for heatmap plots")
+@click.option("--no_plot", is_flag=True,
+              help="Skip plotting geneset significance heatmaps")
+
+def cmd_gprofiler(output_dir, pkl_file, h5ad_file, gene_sets, species, min_intersection, max_intersection, n_hsg, cmap, vmin, vmax, no_plot):
+    """
+    Perform gProfiler gene set analysis of highly-scoring genes from mosaicMPI programs. If a network_integration.pkl file
+    is provided, gProfiler is performed on reprepresentative programs from an integration. If a .h5ad file is provided, gProfiler
+    is performed on all programs.
+    """
+    utils.start_logging()  # allows warning messages to be printed even though logfile hasn't been made yet
+
+    try:
+        from .gprofiler import program_gprofiler, order_genesets
+    except ImportError:
+        logging.error("gprofiler-official is not installed. Please install using:\n\n\t"
+                      "conda install -c bioconda gprofiler-official"
+                      )
+        sys.exit(1)
+
+    # create directory structure, warn if not empty
+    output_dir = os.path.normpath(output_dir)
+    os.makedirs(output_dir, exist_ok=True)
+    if os.listdir(output_dir):
+        logging.warning(f"{output_dir} is not empty. Files may be overwritten.")
+
+    # write to log file
+    utils.start_logging(os.path.join(output_dir, "logfile.txt"))
+
+    if h5ad_file and not pkl_file:
+        # run gProfiler on all programs from a .h5ad file
+        dataset = Dataset.from_h5ad(h5ad_file)
+        programs = dataset.get_programs()
+        
+        logging.info(f"Querying g:Profiler...")
+        result = program_gprofiler(program_df=programs, species=species, n_hsg=n_hsg,
+                                   gene_sets=gene_sets, min_intersection=min_intersection, max_intersection=max_intersection)
+
+        # write background genes
+        with open(os.path.join(output_dir, f"background.txt"), "w") as f:  # gProfiler input format for web version
+            f.write(f">background\n")
+            f.write(" ".join(result.background) + "\n")
+
+        # write query genes
+        with open(os.path.join(output_dir, "query.txt"), "w") as f:   # gProfiler input format for web version
+            for name, querylist in result.query.items():
+                f.write(f">{name}\n")
+                f.write(" ".join(querylist) + "\n")
+
+        # write gProfiler output
+        result.gprofiler_output.to_csv(os.path.join(output_dir, "result.full.txt"), sep="\t")
+        result.summary.to_csv(os.path.join(output_dir, "result.txt"), sep="\t")
+
+        # create p-value heatmaps separately for each k
+        max_k = dataset.adata.uns["kvals"].index.max()
+        for k in tqdm(dataset.adata.uns["kvals"].index, total=max_k, unit="k", desc="Creating heatmaps"):
+            df = result.summary["-log10pval"][k].dropna(how="all").fillna(0)
+            df = order_genesets(df)
+            df.to_csv(os.path.join(output_dir, f"ordered_genesets_k{k}.txt"), sep="\t")
+            if not no_plot:
+                fig, figlegend = plot_geneset_pval_heatmap(df=df)
+                ax = fig.axes[0]
+                ax.set_title("g:Profiler pathways\n" + ",".join(gene_sets))
+                ax.set_xlabel("")
+                ax.set_ylabel("")
+                for _, spine in ax.spines.items():
+                    spine.set_visible(True)
+                    spine.set_color('#aaaaaa') 
+                ax.set_xlabel(f"Program (k={k})")
+                figlegend.savefig(os.path.join(output_dir, f"ordered_genesets_k{k}.legend.pdf"))
+                fig.savefig(os.path.join(output_dir, f"ordered_genesets_k{k}.pdf"))
+                plt.close(fig)
+                plt.close(figlegend)
+
+
+    elif pkl_file and not h5ad_file:
+        # run gProfiler on representative programs for each dataset in a network_integration.pkl file
+        network = Network.from_pkl(pkl_file)
+        rep_programs = network.get_representative_programs()
+        for dataset_name in network.integration.datasets.keys():
+            programs = rep_programs.xs(dataset_name, axis=1, level=1).dropna()
+            programs = programs.sort_index(axis=1, level=0, key=network.get_vectorized_community_sort_key).droplevel(axis=1, level=[1, 2])
+
+            logging.info(f"Querying g:Profiler...")
+            result = program_gprofiler(program_df=programs, species=species, n_hsg=n_hsg,
+                                    gene_sets=gene_sets, min_intersection=min_intersection, max_intersection=max_intersection)
+
+            # write background genes
+            os.makedirs(os.path.join(output_dir, dataset_name), exist_ok=True)
+            with open(os.path.join(output_dir, dataset_name, f"background.txt"), "w") as f:  # gProfiler input format for web version
+                f.write(f">background\n")
+                f.write(" ".join(result.background) + "\n")
+
+            # write query genes
+            with open(os.path.join(output_dir, dataset_name, "query.txt"), "w") as f:   # gProfiler input format for web version
+                for name, querylist in result.query.items():
+                    f.write(f">{name}\n")
+                    f.write(" ".join(querylist) + "\n")
+
+            # write gProfiler output
+            result.gprofiler_output.to_csv(os.path.join(output_dir, dataset_name, "result.full.txt"), sep="\t")
+            result.summary.to_csv(os.path.join(output_dir, dataset_name, "result.txt"), sep="\t")
+            
+            # ordered pathways
+            df = result.summary["-log10pval"].dropna(how="all").fillna(0)
+            df = order_genesets(df)
+            df.to_csv(os.path.join(output_dir, dataset_name, f"ordered_genesets.txt"), sep="\t")
+
+            if not no_plot:
+                fig, figlegend = plot_geneset_pval_heatmap(df=df)
+                ax = fig.axes[0]
+                ax.set_title("g:Profiler pathways\n" + ",".join(gene_sets))
+                ax.set_xlabel("")
+                ax.set_ylabel("")
+                for _, spine in ax.spines.items():
+                    spine.set_visible(True)
+                    spine.set_color('#aaaaaa') 
+                ax.set_xlabel(f"Community\n(Representative Program)")
+                figlegend.savefig(os.path.join(output_dir, dataset_name, f"ordered_genesets.legend.pdf"))
+                fig.savefig(os.path.join(output_dir, dataset_name, f"ordered_genesets.pdf"))
+                plt.close(fig)
+                plt.close(figlegend)
+        
+    else:
+        logging.error("mosaicmpi gprofiler requires either a factorized dataset (.h5ad) file or "
+                      "a network_integration.pkl.gz file to run g:Profiler on programs.")
+        sys.exit(1)
+    logging.info("All tasks completed successfully.")
+
+
 @click.command(name="compare-integrations")
 @click.option('-o', '--output_dir', type=click.Path(file_okay=False), required=True,
     help="Output directory for results")
@@ -1079,8 +1283,10 @@ cli.add_command(cmd_set_parameters)
 cli.add_command(cmd_factorize)
 cli.add_command(cmd_postprocess)
 cli.add_command(cmd_annotated_heatmap)
+cli.add_command(cmd_map_gene_ids)
 cli.add_command(cmd_create_config)
 cli.add_command(cmd_integrate)
 cli.add_command(cmd_ssgsea)
+cli.add_command(cmd_gprofiler)
 cli.add_command(cmd_compare_integrations)
 cli.add_command(cmd_transfer_labels)
