@@ -1,5 +1,5 @@
 
-from . import utils, cnmf, __version__
+from . import utils, cnmf, biomart, __version__
 
 import numpy as np
 import pandas as pd
@@ -9,10 +9,10 @@ import sys
 from typing import Union, Optional, Literal
 from collections.abc import Iterable, Collection
 import os
+from io import StringIO
 from glob import glob
 
-import scipy as sp
-import scanpy as sc
+import scipy.sparse as sp
 import anndata as ad
 import pandas as pd
 import numpy as np
@@ -247,14 +247,199 @@ class Dataset():
         :return: gene list
         :rtype: list
         """
-        return self.adata.uns["gene_list"]
+        return self.adata.var[self.adata.var["selected"]].index
 
 
     @mosaicmpi_version.setter
     def mosaicmpi_version(self, value: bool):
         self.adata.uns["mosaicmpi_version"] = value
 
-        
+    def map_gene_ids(self,
+                        source_species: Literal["hsapiens", "mmusculus", "rnorvegicus", "sscrofa", "dmelanogaster", "drerio", "celegans"],
+                        dest_species: Literal["hsapiens", "mmusculus", "rnorvegicus", "sscrofa", "dmelanogaster", "drerio", "celegans"],
+                        source_ids: Literal["ensembl_gene", "gene_name"],
+                        dest_ids: Literal["ensembl_gene", "gene_name"],
+                        one_to_one: bool = True,
+                        one_to_many: Literal[False, "duplicate", "divide"] = False,
+                        many_to_one: Literal[False, "mean", "sum"] = False,
+                        many_to_many: Literal[False, "mean", "sum"] = False,
+                        unmapped_prefix: str = "unmapped_",
+                        case_sensitive: bool = False,
+                        biomart_url: str = "http://www.ensembl.org:80/biomart/martservice"):
+        """Map the feature IDs in place for a dataset. Mapping occurs from a source to the dest species, and can be gene names or ensembl gene IDs (eg., ENSG..., ENSMUSG...).
+
+        :param source_species: Species name for IDs in the dataset.
+        :type source_species: Literal["hsapiens", "mmusculus", "rnorvegicus", "sscrofa", "dmelanogaster", "drerio", "celegans"]
+        :param dest_species: Species name for IDs after mapping
+        :type dest_species: Literal["hsapiens", "mmusculus", "rnorvegicus", "sscrofa", "dmelanogaster", "drerio", "celegans"]
+        :param source_ids: Whether the IDs are gene names (eg., EGFR), or Ensembl genes (eg., ENSG00000146648)
+        :type source_ids: Literal["ensembl_gene", "gene_name"]
+        :param dest_ids: Whether the IDs should be mapped to gene names (eg., EGFR), or Ensembl genes (eg., ENSG00000146648)
+        :type dest_ids: Literal["ensembl_gene", "gene_name"]
+        :param one_to_one: Whether to map genes that have a one-to-one mapping, defaults to True
+        :type one_to_one: bool, optional
+        :param one_to_many: Whether to map genes that have a one-to-many mapping, defaults to False
+        :type one_to_many: Literal[False, "duplicate", "divide"], optional
+        :param many_to_one: Whether and how to map genes that have a many-to-one mapping, defaults to False
+        :type many_to_one: Literal[False, "mean", "sum"], optional
+        :param many_to_many: Whether and how to map genes that have a many-to-many mapping, defaults to False
+        :type many_to_many: Literal[False, "mean", "sum"], optional
+        :param unmapped_prefix: For unmapped features, prepend this text to their ID, defaults to "unmapped\\_"
+        :type unmapped_prefix: str, optional
+        :param case_sensitive: Case-sensitive ID matching, defaults to False
+        :type case_sensitive: bool, optional
+        :param biomart_url: URL to connect to the Biomart web server, defaults to
+                            "http://www.ensembl.org:80/biomart/martservice"
+        :type biomart_url: str, optional
+        :raises NotImplementedError: for features net yet implemented, including many-to-one and many-to-many gene mappings
+        """
+        logging.info("Downloading gene ID table from Ensembl Biomart")
+        gene_dataset = biomart.BiomartServer(url=biomart_url).datasets[f"{source_species}_gene_ensembl"]
+        if source_species == dest_species:
+            result = gene_dataset.search(params={"attributes": ["external_gene_name",
+                                                                "ensembl_gene_id"]})
+            columns=['source_gene_name','source_ensembl_gene']
+            result = pd.read_csv(StringIO(result.text), sep="\t", header=None, names=columns)
+            source_col = f"source_{source_ids}"
+            dest_col = f"source_{dest_ids}"
+        else:
+            result = gene_dataset.search(params={"attributes": ["external_gene_name",
+                                                            "ensembl_gene_id",
+                                                            f"{dest_species}_homolog_associated_gene_name",
+                                                            f"{dest_species}_homolog_ensembl_gene"]})
+            columns=['source_gene_name','source_ensembl_gene', 'dest_gene_name', 'dest_ensembl_gene']
+            result = pd.read_csv(StringIO(result.text), sep="\t", header=None, names=columns)
+            source_col = f"source_{source_ids}"
+            dest_col = f"dest_{dest_ids}"
+
+        logging.info("Mapping gene IDs")
+
+        if not case_sensitive:
+            result[source_col] = result[source_col].str.casefold()
+
+        id_mapping = result.groupby([source_col, dest_col]).apply(lambda x : x.count()).iloc[:,0]
+        id_mapping = pd.DataFrame(id_mapping.rename("path_counts"))
+        assert id_mapping.index.is_unique
+        source_id_counts = id_mapping.index.get_level_values(0).value_counts()
+        dest_id_counts = id_mapping.index.get_level_values(1).value_counts()
+        id_mapping["source_id_count"] = id_mapping.index.get_level_values(0).map(source_id_counts)
+        id_mapping["dest_id_count"] = id_mapping.index.get_level_values(1).map(dest_id_counts)
+
+        # create ordered lists of the IDs for indexing the anndata object
+        source_id_list = []
+        dest_id_list= []
+        mapping_relationship = []
+
+        # One-to-none: IDs that are not found in the ID mapping table
+        if case_sensitive:
+            id_idx = ~self.adata.var_names.isin(id_mapping.index.get_level_values(0))
+        else:
+            id_idx = ~self.adata.var_names.str.casefold().isin(id_mapping.index.get_level_values(0))
+        src_ids = self.adata.var_names[id_idx]
+        dest_ids = unmapped_prefix + src_ids
+        source_id_list.extend(src_ids)
+        dest_id_list.extend(unmapped_prefix + src_ids)
+        mapping_relationship.extend(len(src_ids) * ["one-to-none"])
+
+        # One-to-one
+        o2o = id_mapping[(id_mapping["source_id_count"] == 1) & (id_mapping["dest_id_count"] == 1)]
+        if case_sensitive:
+            id_idx = self.adata.var_names.isin(o2o.index.get_level_values(0))
+            ids_to_map = self.adata.var_names[id_idx]
+        else:
+            id_idx = self.adata.var_names.str.casefold().isin(o2o.index.get_level_values(0).str.casefold())
+            ids_to_map = self.adata.var_names[id_idx].str.casefold()
+        src_ids = self.adata.var_names[id_idx]
+        if one_to_one:
+            dest_ids = o2o.index.to_frame().reset_index(level=1, drop=True)[dest_col].loc[ids_to_map]
+        else:
+            dest_ids = unmapped_prefix + src_ids
+        source_id_list.extend(src_ids)
+        dest_id_list.extend(dest_ids)
+        mapping_relationship.extend(len(src_ids) * ["one-to-one"])
+
+        # One-to-many
+        o2m = id_mapping[(id_mapping["source_id_count"] == 1) & (id_mapping["dest_id_count"] > 1)]
+        if case_sensitive:
+            id_idx = self.adata.var_names.isin(o2m.index.get_level_values(0))
+            ids_to_map = self.adata.var_names[id_idx]
+        else:
+            id_idx = self.adata.var_names.str.casefold().isin(o2m.index.get_level_values(0).str.casefold())
+            ids_to_map = self.adata.var_names[id_idx].str.casefold()
+        src_ids = self.adata.var_names[id_idx]
+        if one_to_many == "duplicate":
+            # duplicates the values of the existing ID for the new multiple IDs
+            for src_id, id_to_map in zip(src_ids, ids_to_map):
+                dest = o2m.loc[id_to_map].index
+                source_id_list.extend(len(dest) * [src_id])
+                dest_id_list.extend(dest)
+                mapping_relationship.extend(len(dest) * ["one-to-many"])
+        elif one_to_many is False:
+            dest_ids = unmapped_prefix + src_ids
+            source_id_list.extend(src_ids)
+            dest_id_list.extend(dest_ids)
+            mapping_relationship.extend(len(src_ids) * ["one-to-many"])
+        else:
+            raise NotImplementedError(f"`{one_to_many}` is not currently implemented for one-to-many gene mappings.")
+
+        # Many-to-one only
+        m2o = id_mapping[(id_mapping["source_id_count"] > 1) & (id_mapping["dest_id_count"] == 1)]
+        m2m = id_mapping[(id_mapping["source_id_count"] > 1) & (id_mapping["dest_id_count"] > 1)]
+
+        if case_sensitive:
+            id_idx = self.adata.var_names.isin(m2o.index.get_level_values(0)) & ~self.adata.var_names.isin(m2m.index.get_level_values(0))
+            ids_to_map = self.adata.var_names[id_idx]
+        else:
+            id_idx = self.adata.var_names.str.casefold().isin(m2o.index.get_level_values(0).str.casefold()) & ~self.adata.var_names.str.casefold().isin(m2m.index.get_level_values(0).str.casefold())
+            ids_to_map = self.adata.var_names[id_idx].str.casefold()
+        src_ids = self.adata.var_names[id_idx]
+        if many_to_one is False:
+            dest_ids = unmapped_prefix + src_ids
+            source_id_list.extend(src_ids)
+            dest_id_list.extend(dest_ids)
+            mapping_relationship.extend(len(src_ids) * ["many-to-one"])
+        else:
+            raise NotImplementedError(f"`{many_to_one}` is not currently implemented for many-to-one gene mappings.")
+
+        # Many-to-many only
+        if case_sensitive:
+            id_idx = self.adata.var_names.isin(m2m.index.get_level_values(0)) & ~self.adata.var_names.isin(m2o.index.get_level_values(0))
+            ids_to_map = self.adata.var_names[id_idx]
+        else:
+            id_idx = self.adata.var_names.str.casefold().isin(m2m.index.get_level_values(0).str.casefold()) & ~self.adata.var_names.str.casefold().isin(m2o.index.get_level_values(0).str.casefold())
+            ids_to_map = self.adata.var_names[id_idx].str.casefold()
+        src_ids = self.adata.var_names[id_idx]
+        if many_to_many is False:
+            dest_ids = unmapped_prefix + src_ids
+            source_id_list.extend(src_ids)
+            dest_id_list.extend(dest_ids)
+            mapping_relationship.extend(len(src_ids) * ["many-to-many"])
+        else:
+            raise NotImplementedError(f"`{many_to_many}` is not currently implemented for many-to-many gene mappings.")
+
+        # IDs with both many-to-many and many-to-one relationship
+        if case_sensitive:
+            id_idx = self.adata.var_names.isin(m2m.index.get_level_values(0)) & self.adata.var_names.isin(m2o.index.get_level_values(0))
+            ids_to_map = self.adata.var_names[id_idx]
+        else:
+            id_idx = self.adata.var_names.str.casefold().isin(m2m.index.get_level_values(0).str.casefold()) & self.adata.var_names.str.casefold().isin(m2o.index.get_level_values(0).str.casefold())
+            ids_to_map = self.adata.var_names[id_idx].str.casefold()
+        src_ids = self.adata.var_names[id_idx]
+        dest_ids = unmapped_prefix + src_ids
+        source_id_list.extend(src_ids)
+        dest_id_list.extend(dest_ids)
+        mapping_relationship.extend(len(src_ids) * ["many-to-many; many-to-one"])
+    
+        # create new anndata object
+        new_adata = self.adata[:, source_id_list]
+        new_adata.var_names = dest_id_list
+        new_adata.var["source_id"] = source_id_list
+        new_adata.var["mapping_relationship"] = mapping_relationship
+        new_adata.var["mapping_relationship"] = new_adata.var["mapping_relationship"].astype("category")
+        self.adata = new_adata
+        self.append_to_history(f"Mapped feature IDs. source species={source_species}, source_ids={source_ids}, dest_species={dest_species}, dest_ids={dest_ids}, "
+                               f"one_to_one={one_to_one}, one_to_many={one_to_many}, many_to_one={many_to_one}, many_to_many={many_to_many}, unmapped_prefix={unmapped_prefix}")
+
     def update_obs(self, obs):
         """Update the observation metadata with a new metadata matrix
 
@@ -313,9 +498,25 @@ class Dataset():
         if normalized and not self.is_normalized:
             df = df.div(df.sum(axis=1), axis=0) * 1e6  # TPM normalization
         return df
+
+    def remove_unfactorizable_observations(self):
+        """Removes observations with all zeros from the data matrix.
+        """
+        df = self.to_df(normalized=False)
+        # Check for observations with all zeros
+        obs_zeros = (df.T == 0).all()
+        n_obs_zeros = obs_zeros.sum()
+        logging.info(f"{n_obs_zeros} of {self.adata.n_obs} observations are all zeros.")
+        if n_obs_zeros:
+            logging.warning(f"Subsetting observations to those with at least 1 positive value.")
+            msg_string = repr(obs_zeros[obs_zeros].index.to_list())
+            self.append_to_history(f"Removing observations with only zeros: {msg_string}")
         
+        self.adata = self.adata[~obs_zeros,:].copy()
+
+
     def remove_unfactorizable_features(self):
-        """Removes genes with missing values or zero variance from the data matrix.
+        """Removes features with missing values or zero variance from the data matrix.
         """
         df = self.to_df(normalized=False)
         # Check for features with missing values
@@ -325,15 +526,16 @@ class Dataset():
         logging.info(f"{n_missing} of {self.adata.n_vars} features are missing values.")
         if n_missing:
             logging.warning(f"Subsetting features to those with no missing values.")
-            self.append_to_history(f"Removing features with missing values: {genes_with_missingvalues}")
+            msg_string = repr(genes_with_missingvalues[genes_with_missingvalues].index.to_list())
+            self.append_to_history(f"Removing features with missing values: {msg_string}")
         # Check for genes with zero variance
         zerovargenes = (df.var() == 0)
         n_zerovar = zerovargenes.sum()
         logging.info(f"{n_zerovar} of {self.adata.n_vars} features have a variance of zero.")
         if n_zerovar:
             logging.warning(f"Subsetting features to those with nonzero variance.")
-            self.append_to_history(f"Removing features with zero variance: {zerovargenes}")
-        
+            msg_string = repr(zerovargenes[zerovargenes].index.to_list())
+            self.append_to_history(f"Removing features with zero variance: {msg_string}")
         genes_to_keep = (~genes_with_missingvalues) & (~zerovargenes)
         self.adata = self.adata[:,genes_to_keep].copy()
 
@@ -395,7 +597,7 @@ class Dataset():
             - 'distance' : weight points by the inverse of their distance.
             in this case, closer neighbors of a query point will have a
             greater influence than neighbors which are further away.
-        :type weights: Literal[&quot;distance&quot;, &quot;uniform&quot;], optional
+        :type weights: Literal["distance", "uniform"], optional
         :param cross_validate: perform k-fold cross-validation, defaults to True
         :type cross_validate: bool, optional
         :param n_folds: number of folds for k-fold cross-validation, defaults to 100
@@ -503,7 +705,7 @@ class Dataset():
         :param min_mean: minimum gene expression for genes to be counted as overdispersed, defaults to 0
         :type min_mean: int, optional
         """
-        self.adata.var["selected"] = self.adata.var.index.isin(genes) & self.adata.var["mean_counts"] >= min_mean
+        self.adata.var["selected"] = self.adata.var.index.isin(genes) & (self.adata.var["mean_counts"] >= min_mean)
         self.adata.uns["odg"]["overdispersion_metric"] = ""
         self.adata.uns["odg"]["min_mean"] = min_mean
         self.adata.uns["odg"]["min_score"] = ""
@@ -624,10 +826,11 @@ class Dataset():
         # Subset out high-variance genes
         norm_counts = self.adata[:, overdispersed_genes]
         ## Scale genes to unit variance
-        if sp.sparse.issparse(tpm.X):
-            sc.pp.scale(norm_counts, zero_center=False)
-            if np.isnan(norm_counts.X.data).sum() > 0:
-                raise ValueError('NaNs in normalized counts matrix')                       
+        if sp.issparse(tpm.X):
+            raise NotImplementedError("AnnDatas with sparse matrices are not supported by mosaicMPI yet.")
+            # sc.pp.scale(norm_counts, zero_center=False)
+            # if np.isnan(norm_counts.X.data).sum() > 0:
+            #     raise ValueError('NaNs in normalized counts matrix')                       
         else:
             norm_counts.X /= norm_counts.X.std(axis=0, ddof=1)
             if np.isnan(norm_counts.X).sum().sum() > 0:
@@ -639,7 +842,7 @@ class Dataset():
         if norm_counts.X.dtype != np.float64:
             norm_counts.X = norm_counts.X.astype(np.float64)
 
-        sc.write(cnmf_obj.paths['normalized_counts'], norm_counts)
+        norm_counts.write_h5ad(cnmf_obj.paths['normalized_counts'])
 
         # save parameters for factorization step
         cnmf_obj.save_nmf_iter_params(*cnmf_obj.get_nmf_iter_params(ks=kvals, n_iter=n_iter, random_state_seed=seed, beta_loss=beta_loss))
@@ -770,7 +973,7 @@ class Dataset():
                  type="cnmf_gep_score"
                  ) -> pd.DataFrame:
         """
-        Get GEPs.
+        Get feature scores for programs.
 
         :param k: If an integer or list of integers, returns GEPs only for specified ranks. Otherwise, returns GEPs from all ranks. Defaults to None
         :type k: Union[int, Iterable], optional
@@ -782,9 +985,12 @@ class Dataset():
         df = self.adata.varm[type].copy()
         df.columns = pd.MultiIndex.from_tuples(df.columns.str.split(".").to_list())
         df.columns = df.columns.set_levels([l.astype("int") for l in df.columns.levels])
-        if isinstance(k, (int, Iterable)):
+        if isinstance(k, int):
             df = df.loc[:, k]
-        df = df.sort_index(axis=1)
+            df = df.rename_axis(columns=["program"]).sort_index(axis=1)
+        elif isinstance(k, Iterable):
+            df = df.loc[:, k]
+            df = df.rename_axis(columns=["k", "program"]).sort_index(axis=1)
         return df
     
     def get_metadata_df(self,
@@ -812,7 +1018,7 @@ class Dataset():
             raise ValueError(f"{unexplained_col_str} metadata columns have unrecognized dtypes.")
         df = self.adata.obs.select_dtypes(include=dtypes)
         df = df.dropna(axis=1, how="all")
-        df = df.replace("nan", np.NaN)
+        df = df.replace("nan", np.NaN)  #TODO: Address error: /home/tbverhey/miniconda3/envs/mosaicmpi/lib/python3.11/site-packages/mosaicmpi/dataset.py:1021: FutureWarning: The behavior of Series.replace (and DataFrame.replace) with CategoricalDtype is deprecated. In a future version, replace will only be used for cases that preserve the categories. To change the categories, use ser.cat.rename_categories instead.
         return df
     
     def get_category_overrepresentation(self,
@@ -820,7 +1026,7 @@ class Dataset():
                                         truncate_negative: bool = True,
                                         subset_categories: Collection[str] = None
                                         ) -> pd.DataFrame:
-        """Calculate Pearson residual of chi-squared test, associating GEPs for each rank (k) to categories of samples/observations. By default, truncates negative values.
+        """Calculate Pearson residual of chi-squared test, associating programs for each rank (k) to categories of samples/observations. By default, truncates negative values.
 
         :param layer: name of categorical data layer
         :type layer: str
