@@ -58,7 +58,7 @@ class Integration():
             kvals = dataset.adata.uns["kvals"].copy()
             kvals["cNMF result"] = True
             combined[dataset_name] = kvals
-        self.k_table = pd.concat(combined, axis=1).rename_axis("k", axis=0)
+        self.k_table = pd.concat(combined, axis=1).rename_axis(index="k")
         # compute correlations
         self.compute_corr(method=self.corr_method)
         # rank-reduction for highly autocorrelated GEPs 
@@ -313,7 +313,7 @@ class Integration():
                             if (dataset_row, dataset_col) in thresholds.index:
                                 min_corr = thresholds.loc[(dataset_row, dataset_col)]
                                 mask = df_filt.loc[dataset_row, dataset_col] < min_corr
-                                df_filt.loc[dataset_row, dataset_col][~mask] = np.NaN
+                                df_filt.loc[dataset_row, dataset_col][~mask] = np.NaN   # TODO: Not compatible with Pandas 3.0 because of setting with chained indexing.
                                 df_filt.dropna(axis=0).dropna(axis=1)
                 else:
                     df_filt = df
@@ -323,8 +323,7 @@ class Integration():
                     results[dataset_name] = subdf.shape[0]
                 nodetable[(node_filter, edge_filter)] = results
 
-        nodetable = pd.DataFrame(nodetable)
-        nodetable.columns.rename(["Node filter", "Edge Filter"], inplace=True)
+        nodetable = pd.DataFrame(nodetable).rename_axis(columns=["Node filter", "Edge Filter"])
         return nodetable
     
     def get_metadata_df(self,
